##!<#PERLEXE#>
#
# This script is used to bundle an installed Ganymede schema kit back
# up for distribution, generating a tar file with all of the files
# in the schema kit reverted back to their pre-installed state.
#
# This script can take a single parameter, the name of an archive
# to create.
#
# Jonathan Abbey
# 25 November 2009
#
###############################################################################

use File::Copy;
use Cwd;
use English;

$schemaname = "userKit";
$debug = 0;

# Steps:
#
# 0. Ask for the name of the tar file to be generated, if it isn't
# provided on the command line. This can include both a directory path
# and a filename, with a .tar or .tar.gz extension.  The filename of the
# tar file will be taken as the name to use for the top-level directory
# in the generated archive

$rootdir = &resolve(cwd(),$ENV{GROOTDIR});

if (defined $ARGV[0]) {
  $archivename = &canonify($ARGV[0]);
}

if (!(defined $archivename && (($archivename =~ /\.tar$/) || ($archivename =~ /\.tar\.gz$/)))) {

  $archivename = &resolve(cwd(), "$rootdir/..")."/$schemaname.tar.gz";

  if (cwd() eq $rootdir) {
    $displayname = "../$schemaname.tar.gz";
  } else {
    $displayname = $archivename;
  }

  $success = 0;

  while (!$success) {
    print "Enter the name of the archive to create: [$displayname] ";
    $inp = <>;
    chop $inp;

    if ($inp ne "") {
      $archivename = $inp;
      
      $archivename = &canonify($archivename);
    }
    
    if (-f $archivename) {
      print "Error, $archivename already exists\n";
      $success = 0;
      next;
    }

    if (!(($archivename =~ /\.tar$/) || ($archivename =~ /\.tar\.gz$/))) {
      print "Error, $archivename is not a valid tar archive name\n";
      $success = 0;
      next;
    }

    $dirpath = &dirname($archivename);

    if (!-d $dirpath || !(-w _)) {
      print "Error, $dirpath is not a valid directory to create the archive in\n";
      $success = 0;
      next;
    }

    $success = 1;
  }
}

print "yay!  $archivename\n" if $debug;

# 1. Create a directory under /tmp, with a reasonably random name.. give
# this dir permissions 700 so we can use it as a scratch pad, creating
# a new archive to tar within

$tmpdirname = "/tmp/$schemaname"."_pack";

while (-f $tmpdirname) {
  $tmpdirname .= "A";
}

&makedir($tmpdirname, 0700);

# and now within that dir, make a directory with the actual archive
# name, minus the .tar or .tar.gz extensions

@comps = split /\//, $archivename;
$packagename = pop(@comps);

if ($packagename =~ /\.tar\.gz$/) {
  $basename = $packagename;
  $basename =~ s/\.tar\.gz$//;
} elsif ($packagename =~ /\.tar$/) {
  $basename = $packagename;
  $basename =~ s/\.tar$//;
} else {
  die "Couldn't parse packagename $packagename";
}

# we want the directory extracted from the tar
# file to have moderate permissions

&makedir("$tmpdirname/$basename", 0750);

if ($debug) {
  print "package name = $packagename\n";
  print "base name = $basename\n";
  print "tmpdir = $tmpdirname\n";
}

# 2. Copy files and directories from the schema kit installation location
# to this temp directory, preserving permissions, rewriting the #!
# line for perl scripts back to ##!<#PERLEXE#>, and removing config.sh
# and rebuild from the src directory.  This script will have to enumerate
# the files to be copied outside of src and scripts.  Generally, things
# like input and output are simply not copied.

&copydir("$rootdir/src", "$tmpdirname/$basename/src", 
	 "config.sh", "rebuild", "\\.class\$", "\\.git");
&copydir("$rootdir/scripts", "$tmpdirname/$basename/scripts",
	 "\\.git");
&copydir("$rootdir/Modules", "$tmpdirname/$basename/Modules",
	 "^Makefile\$", "\\.git");
&copydir("$rootdir/NTscripts", "$tmpdirname/$basename/NTscripts",
	 "\\.git");

@filesToCopy = ("custom.jar", "schema.xml", "makeArchive", "README", "INSTALL", "COPYRIGHT", "CHANGES");

foreach $file (@filesToCopy) {
  if (-f "$rootdir/$file") {
    $mode = (stat("$rootdir/$file"))[2] & 0777;
    copy ("$rootdir/$file", "$tmpdirname/$basename/$file");
    chmod($mode, "$tmpdirname/$basename/$file");
  }
}

# 3. Copy the installKit.in file from the scripts directory to the top-level,
# so when they unpack the archive they'll find an installKit script ready
# to go.  We do this because installKit itself doesn't install a root-level
# copy of itself into the server directory

if (-f "$rootdir/scripts/installKit.in") {
  $mode = (stat("$rootdir/scripts/installKit.in"))[2] & 0777;
  copy ("$rootdir/scripts/installKit.in", "$tmpdirname/$basename/installKit");
  chmod($mode, "$tmpdirname/$basename/installKit");
} else {
  print "Warning, couldn't find $rootdir/scripts/installKit.in to copy to archive\n";
}

# 4. Generate a tar file out of the contents of the temporary directory

chdir ($tmpdirname);

if ($archivename =~ /\.tar$/) {
  system "tar cf $archivename $basename";
} elsif ($archivename =~ /\.tar\.gz$/) {
  $tarfile = $archivename;
  $tarfile =~ s/\.tar\.gz$/\.tar/;
  system "tar cf $tarfile $basename";
  system "gzip $tarfile";
}

# 5. Remove the temporary directory

# we need to get our current working directory out of the way before
# deleting our temp dir

chdir ("/");

system "rm -rf $tmpdirname";

exit;

######################################################################### 
#
#                                                                 copydir
#
# input: 1) a directory to copy from
#        2) directory target
#        3) an optional array of files or directory names not to copy
#
######################################################################### 
sub copydir{
  my ($source, $target, @skip) = @_;
  my (@dirs, $file, $mode);

  &removelastslash($source);
  &removelastslash($target);

  if (!-e $target) {
    &makedir($target, 0750);
  }

  opendir SOURCE, $source || die "Failure in copydir";
  @dirs = readdir SOURCE;
  closedir SOURCE;

  FILELOOP: foreach $file (@dirs) {
      if (($file eq ".") || ($file eq "..")) {
	next;
      }

      foreach $pat (@skip) {
	if ($file =~ /$pat/) {
	  print "Skipping $file because of pattern '$pat'\n" if $debug;
	  next FILELOOP;
	}
      }
      
      if (-d "$source/$file") {
	&copydir("$source/$file", "$target/$file", (@skip)); #recurse
      } else {
	$mode = (stat("$source/$file"))[2] & 0777;
	
	if (($file =~ /\.pl$/) || ($file =~ /\.perl$/)) {
	  
	  # we've got a perl script, copy it and fix the perl invocation line back to
	  # ##!<#PERLEXE#> along the way
	  
	  open(COPYFILEIN, "<$source/$file") || die ("Can't read $source/$file");
	  open(COPYFILEOUT, ">$target/$file") || die("Can't create the $target/file");
	  
	  while (<COPYFILEIN>){
	    s/^\#!.*/\#\#!<\#PERLEXE\#>/;
	    print COPYFILEOUT $_;
	  }
	  
	  close(COPYFILEOUT);
	  close(COPYFILEIN);
	} else {
	  copy("$source/$file", "$target/$file");
	}
	chmod($mode, "$target/$file");
      }
    }
}

######################################################################### 
#
#                                                                 makedir
#
# input: 1) a directory to make
#        2) octal chmod bits
#
######################################################################### 
sub makedir{
  my ($dirpath, $chmod) = @_;

  if (!-e $dirpath) {
    mkdir ($dirpath, $chmod) or die("*Couldn't make $dirpath*");
  }
}

######################################################################### 
#
#                                                         removelastslash
#
# input: a pathname to test
#
# this function will remove a trailing slash from the directory name 
# input
#
######################################################################### 
sub removelastslash{
  if ($_[0] =~ /\/$/) {
    chop $_[0];
  }

  return $_[0];
}

######################################################################### 
#
#                                                                 dirname
#
# input: 1) a filename path
#
# return: the directory containing the filename given
#
######################################################################### 
sub dirname{
  my ($dirpath) = @_;
  my (@comps);

  @comps = split(/\//, $dirpath);
  pop @comps;
  return join('/', @comps);
}

#########################################################################
#
#                                                                 resolve
#
# input: $dir - absolute pathname of current directory
#        $link - string containing the readlink() results for a 
#                symbolic link in $dir to be processed
#
# returns: absolute pathname of the target of the symbolic link
#
#########################################################################
sub resolve{
    my($dir, $link) = @_;
    my(@alinkp, $d, $alinkp);

    # make array representations of
    # the current directory and symbolic link

    # if we have a leading / in our $dir or $link,
    # we'll need to shift to get rid of the leading
    # empty array element
  
    @dirp=split(/\//, $dir);
    shift(@dirp) if (!($dirp[0]));
    @linkp=split(/\//, $link);
    shift(@linkp) if (!($linkp[0]));

    # @alinkp is an array that we will build to contain the absolute link
    # target pathname.  If the link does not begin with a /, it is a relative link,
    # and we need to place our current directory into the @alinkp array.  

    if ($link !~ /^\//) {
	@alinkp=@dirp;
    }

    # modify the @alinkp array according
    # to each path component of the @linkp array
    # (an array representation of the symbolic link
    # given to us), to arrive at the ultimate absolute
    # pathname of the symbolic link

    $d = shift(@linkp);

    while ($d) {
	if ($d eq "..") {
	    pop(@alinkp);
	}
	elsif ($d ne "."){		       
	    push(@alinkp, $d);
	}
	$d=shift(@linkp);
    }

    $"='/';

    # perl functions return the value of the last expression
    # in the subroutine

    $alinkp="/@alinkp";
}

#########################################################################
#
#                                                                canonify
#
# input: $pathname - a directory or filename path, may be relative
# or use ~ to represent the current user's home directory
#
# returns: absolute location of the givevn pathname
#
#########################################################################
sub canonify {
  my ($pathname) = @_;
  my ($homedir);

  if (!(($pathname =~ /^\//) || ($pathname =~ /^~\//))) {
    $pathname = &resolve(cwd(), $pathname);
  }

  if ($pathname =~ /^~\//) {
    $homedir = $ENV{'HOME'};
    $pathname =~ s/\~\//$homedir\//g;
  }

  return $pathname;
}
