/*

   BSDBuilderTask.java

   This class is intended to dump the Ganymede datastore to the
   BSD passwd, master.passwd, and group files.
   
   Created: 30 July 1998

   Last Mod Date: $Date$
   Last Revision Changed: $Rev$
   Last Changed By: $Author$
   SVN URL: $HeadURL$

   Module By: Jonathan Abbey, jonabbey@arlut.utexas.edu

   -----------------------------------------------------------------------
	    
   Ganymede Directory Management System

   Copyright (C) 1996-2005
   The University of Texas at Austin.

   Contact information

   Author Email: ganymede_author@arlut.utexas.edu
   Email mailing list: ganymede@arlut.utexas.edu

   US Mail:

   Computer Science Division
   Applied Research Laboratories
   The University of Texas at Austin
   PO Box 8029, Austin TX 78713-8029

   Telephone: (512) 835-3200

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA

*/

package arlut.csd.ganymede.userKit;

import arlut.csd.ganymede.server.*;
import arlut.csd.ganymede.common.*;
import arlut.csd.Util.PathComplete;
import arlut.csd.Util.FileOps;

import java.util.*;
import java.text.*;
import java.io.*;

/*------------------------------------------------------------------------------
                                                                           class
                                                                  BSDBuilderTask

------------------------------------------------------------------------------*/

/**
 *
 * This class is intended to dump the Ganymede datastore to the
 * BSD passwd, master.passwd, and group files.
 *
 * @author Jonathan Abbey jonabbey@arlut.utexas.edu
 *
 */

public class BSDBuilderTask extends GanymedeBuilderTask {

  private String buildScript = null;
  private String passwdFile = null;
  private String shadowFile = null;
  private String groupFile = null;
  private boolean useMD5 = false;

  // ---

  private Date now = null;
  private boolean backedup = false;

  /* -- */

  /**
   * <p>The constructor for GanymedeBuilderTask subclasses takes an
   * Invid for the DBObject in the Ganymede data store that defines
   * the task.  The taskDefObjInvid member variable is comes from the
   * GanymedeBuilderTask that we are inheriting from.  Setting it
   * in this constructor allows us to use the getOptionValue() method to read
   * our configuration strings for this task.</p>
   */

  public BSDBuilderTask(Invid _taskObjInvid)
  {
    // set the taskDefObjInvid in GanymedeBuilderTask so
    // we can look up option strings

    taskDefObjInvid = _taskObjInvid;
  }

  /**
   *
   * This method runs with a dumpLock obtained for the builder task.
   *
   * Code run in builderPhase1() can call enumerateObjects() and
   * baseChanged().
   *
   * @return true if builderPhase1 made changes necessitating the
   * execution of builderPhase2.
   *
   */

  public boolean builderPhase1()
  {
    PrintWriter out;
    boolean result = false;

    /* -- */

    // initialize operational variables for the new run

    backedup = false;
    now = null;

    // now see what our current option values are

    useMD5 = isOptionSet("useMD5");

    passwdFile = getOptionValue("passwdFile");

    if (passwdFile == null)
      {
	Ganymede.debug("BSDBuilderTask: error, no passwdFile specified");
      }

    shadowFile = getOptionValue("shadowFile");

    groupFile = getOptionValue("groupFile");

    if (groupFile == null)
      {
	Ganymede.debug("BSDBuilderTask: error, no groupFile specified");
      }

    // passwd

    if (baseChanged(SchemaConstants.UserBase))
      {
	Ganymede.debug("BSDBuilderTask: Need to build user map");

	if (shadowFile != null)
	  {
	    out = null;

	    try
	      {
		out = openOutFile(shadowFile, "bsd");
	      }
	    catch (IOException ex)
	      {
		ex.printStackTrace();
		Ganymede.debug("BSDBuilderTask.builderPhase1(): couldn't open " + shadowFile + ": " + ex);
	      }
	    
	    if (out != null)
	      {
		try
		  {
		    DBObject user;
		    Enumeration users = enumerateObjects(SchemaConstants.UserBase);
		    
		    while (users.hasMoreElements())
		      {
			user = (DBObject) users.nextElement();
			
			writeMasterUserLine(user, out);
		      }
		  }
		finally
		  {
		    out.close();
		    out = null;
		  }
	      }

	    result = true;
	  }

	if (passwdFile != null)
	  {
	    boolean assumeShadow = (shadowFile != null);
	    out = null;

	    try
	      {
		out = openOutFile(passwdFile, "bsd");
	      }
	    catch (IOException ex)
	      {
		ex.printStackTrace();
		Ganymede.debug("BSDBuilderTask.builderPhase1(): couldn't open " + passwdFile + ": " + ex);
	      }
	    
	    if (out != null)
	      {
		DBObject user;
		Enumeration users = enumerateObjects(SchemaConstants.UserBase);
		
		while (users.hasMoreElements())
		  {
		    user = (DBObject) users.nextElement();
		    
		    writeUserLine(user, out, assumeShadow);
		  }
		
		out.close();
	      }

	    result = true;
	  }
      }

    // group

    if (groupFile != null && baseChanged((short) 257))
      {
	Ganymede.debug("BSDBuilderTask: Need to build group map");

	out = null;

	try
	  {
	    out = openOutFile(groupFile, "bsd");
	  }
	catch (IOException ex)
	  {
	    ex.printStackTrace();
	    Ganymede.debug("BSDBuilderTask.builderPhase1(): couldn't open " + groupFile + ": " + ex);
	  }
	
	if (out != null)
	  {
	    DBObject group;
	    Enumeration groups = enumerateObjects((short) 257);

	    while (groups.hasMoreElements())
	      {
		group = (DBObject) groups.nextElement();

		writeGroupLine(group, out);
	      }

	    out.close();
	  }

	result = true;
      }

    return result;
  }

  /**
   *
   * This method runs after this task's dumpLock has been
   * relinquished.  This method is intended to be used to finish off a
   * build process by running (probably external) code that does not
   * require direct access to the database.
   *
   * builderPhase2 is only run if builderPhase1 returns true.
   *
   */

  public boolean builderPhase2()
  {
    File
      file;

    /* -- */

    buildScript = getOptionValue("buildScript");

    if (buildScript == null || buildScript.equals(""))
      {
	Ganymede.debug("BSDBuilderTask: no buildScript defined in task object's option string vector");
	Ganymede.debug("                not running any external update script.");
	return false;
      }

    file = new File(buildScript);

    if (file.exists())
      {
	Ganymede.debug("BSDBuilderTask builderPhase2 running");

	try
	  {
	    FileOps.runProcess(buildScript);
	  }
	catch (IOException ex)
	  {
	    Ganymede.debug("Couldn't exec buildScript (" + buildScript + ") due to IOException: " + ex);
	  }
	catch (InterruptedException ex)
	  {
	    Ganymede.debug("Failure during exec of buildScript (" + buildScript + "): " + ex);
	  }
      }
    else
      {
	Ganymede.debug(buildScript + " doesn't exist, not running external BSD build script");
      }

    Ganymede.debug("BSDBuilderTask builderPhase2 completed");

    return true;
  }

  // ***
  //
  // The following private methods are used to support the BSD builder logic.
  //
  // ***

  /**
   *
   * This method writes out a line to the passwd BSD source file.
   *
   * The lines in this file look like the following.
   *
   * broccol:*:12003:12010:Jonathan Abbey,S321 CSD,3199,8343915:/home/broccol:/bin/tcsh
   *
   * @param object An object from the Ganymede user object base
   * @param writer The destination for this user line
   *
   */

  private void writeUserLine(DBObject object, PrintWriter writer, boolean assumeShadow)
  {
    String username;
    int uid;
    int gid;
    String name;
    String room;
    String officePhone;
    String homePhone;
    String directory;
    String shell;

    Vector invids;
    Invid groupInvid;
    DBObject group;

    StringBuffer result = new StringBuffer();

    /* -- */

    username = (String) object.getFieldValueLocal(SchemaConstants.UserUserName);

    uid = ((Integer) object.getFieldValueLocal(userSchema.UID)).intValue();

    // get the gid
    
    groupInvid = (Invid) object.getFieldValueLocal(userSchema.HOMEGROUP); // home group

    if (groupInvid == null)
      {
	System.err.println("BSDBuilderTask.writeUserLine(): null gid for user " + username);
	gid = -1;
      }
    else
      {
	group = getObject(groupInvid);

	if (group == null)
	  {
	    System.err.println("BSDBuilderTask.writeUserLine(): couldn't get reference to home group");

	    gid = -1;
	  }
	else
	  {
	    Integer gidInt = (Integer) group.getFieldValueLocal(groupSchema.GID);

	    if (gidInt == null)
	      {
		System.err.println("BSDBuilderTask.writeUserLine(): couldn't get gid value");

		gid = -1;
	      }
	    else
	      {
		gid = gidInt.intValue();
	      }
	  }
      }

    name = (String) object.getFieldValueLocal(userSchema.FULLNAME);
    room = (String) object.getFieldValueLocal(userSchema.ROOM);
    officePhone = (String) object.getFieldValueLocal(userSchema.OFFICEPHONE);
    homePhone = (String) object.getFieldValueLocal(userSchema.HOMEPHONE);
    directory = (String) object.getFieldValueLocal(userSchema.HOMEDIR);
    shell = (String) object.getFieldValueLocal(userSchema.LOGINSHELL);

    // now build our output line

    result.append(username);

    if (assumeShadow)
      {
	result.append(":*:");	// don't write out the passwd, assume shadow use.
      }
    else
      {
	String cryptedPass = null;
	PasswordDBField passField = (PasswordDBField) object.getField(SchemaConstants.UserPassword);

	if (passField != null)
	  {
	    if (useMD5)
	      {
		cryptedPass = passField.getMD5CryptText();
	      }
	
	    // if the Ganymede server doesn't have an MD5 password for
	    // this user, go ahead and devolve to the crypt() password if
	    // available.  This might have been set by rpcpass and the NIS
	    // passwd daemon.

	    if (cryptedPass == null)
	      {
		cryptedPass = passField.getUNIXCryptText();
	      }
	  }

	if (cryptedPass == null)
	  {
	    result.append(":*:");
	  }
	else
	  {
	    result.append(":");
	    result.append(cryptedPass);
	    result.append(":");
	  }
      }

    result.append(Integer.toString(uid));
    result.append(":");
    result.append(Integer.toString(gid));
    result.append(":");
    result.append(name);

    if (room != null && !room.equals(""))
      {
	result.append(",");
	result.append(room);
      }

    if (officePhone != null && !officePhone.equals(""))
      {
	result.append(",");
	result.append(officePhone);
      }

    if (homePhone != null && !homePhone.equals(""))
      {
	result.append(",");
	result.append(homePhone);
      }

    result.append(":");
    result.append(directory);
    result.append(":");
    result.append(shell);

    if (result.length() > 1024)
      {
	System.err.println("BSDBuilderTask.writeUserLine(): Warning!  user " + 
			   username + " overflows the BSD line length!");
      }

    writer.println(result.toString());
  }

  /**
   *
   * This method writes out a line to the master.passwd BSD source file.
   *
   * The lines in this file look like the following.
   *
   * broccol:393T6k3e/9/w2:12003:12010:user:0:0:Jonathan Abbey,S321 CSD,3199,8343915:/home/broccol:/bin/tcsh
   *
   * @param object An object from the Ganymede user object base
   * @param writer The destination for this user line
   *
   */

  private void writeMasterUserLine(DBObject object, PrintWriter writer)
  {
    String username;
    String cryptedPass = null;
    int uid;
    int gid;
    String name;
    String room;
    String officePhone;
    String homePhone;
    String directory;
    String shell;
    int lastchange;
    int expire;

    PasswordDBField passField;
    Vector invids;
    Invid groupInvid;
    DBObject group;

    StringBuffer result = new StringBuffer();

    /* -- */

    username = (String) object.getFieldValueLocal(SchemaConstants.UserUserName);

    passField = (PasswordDBField) object.getField(SchemaConstants.UserPassword);

    if (passField != null)
      {
	if (useMD5)
	  {
	    cryptedPass = passField.getMD5CryptText();
	  }

	// if the Ganymede server doesn't have an MD5 password for
	// this user, go ahead and devolve to the crypt() password if
	// available.  This might have been set by rpcpass and the NIS
	// passwd daemon.

	if (cryptedPass == null)
	  {
	    cryptedPass = passField.getUNIXCryptText();
	  }
      }
    else
      {
	Ganymede.debug("BSDBuilderTask.writeMasterUserLine(): null password for user " + username);
	cryptedPass = "**Nopass**";
      }

    uid = ((Integer) object.getFieldValueLocal(userSchema.UID)).intValue();

    // get the gid
    
    groupInvid = (Invid) object.getFieldValueLocal(userSchema.HOMEGROUP); // home group

    if (groupInvid == null)
      {
	Ganymede.debug("BSDBuilderTask.writeMasterUserLine(): null gid for user " + username);
	gid = -1;
      }
    else
      {
	group = getObject(groupInvid);

	if (group == null)
	  {
	    Ganymede.debug("BSDBuilderTask.writeMasterUserLine(): couldn't get reference to home group");

	    gid = -1;
	  }
	else
	  {
	    Integer gidInt = (Integer) group.getFieldValueLocal(groupSchema.GID);

	    if (gidInt == null)
	      {
		Ganymede.debug("BSDBuilderTask.writeMasterUserLine(): couldn't get gid value");

		gid = -1;
	      }
	    else
	      {
		gid = gidInt.intValue();
	      }
	  }
      }

    name = (String) object.getFieldValueLocal(userSchema.FULLNAME);
    room = (String) object.getFieldValueLocal(userSchema.ROOM);
    officePhone = (String) object.getFieldValueLocal(userSchema.OFFICEPHONE);
    homePhone = (String) object.getFieldValueLocal(userSchema.HOMEPHONE);
    directory = (String) object.getFieldValueLocal(userSchema.HOMEDIR);
    shell = (String) object.getFieldValueLocal(userSchema.LOGINSHELL);

    // now build our output line

    result.append(username);
    result.append(":");
    result.append(cryptedPass);
    result.append(":");
    result.append(Integer.toString(uid));
    result.append(":");
    result.append(Integer.toString(gid));
    result.append(":");

    // here is where we would put the bsd classification
    // token if we were supporting classification with the
    // user kit

    result.append(":0:0:");
    result.append(name);

    if (room != null && !room.equals(""))
      {
	result.append(",");
	result.append(room);
      }

    if (homePhone != null && !officePhone.equals(""))
      {
	result.append(",");
	result.append(officePhone);
      }

    if (homePhone != null && !homePhone.equals(""))
      {
	result.append(",");
	result.append(homePhone);
      }

    result.append(":");
    result.append(directory);
    result.append(":");
    result.append(shell);

    if (result.length() > 1024)
      {
	Ganymede.debug("BSDBuilderTask.writeMasterUserLine(): Warning!  user " + 
		       username + " overflows the BSD line length!");
      }

    writer.println(result.toString());
  }

  /**
   *
   * This method writes out a line to the passwd BSD source file.
   *
   * The lines in this file look like the following.
   *
   * adgacc:ZzZz:4015:hammp,jgeorge,dd,doodle,dhoss,corbett,monk
   *
   * @param object An object from the Ganymede user object base
   * @param writer The destination for this user line
   *
   */

  private void writeGroupLine(DBObject object, PrintWriter writer)
  {
    String groupname;
    String pass;
    int gid;
    Vector users = new Vector();

    PasswordDBField passField;
    Vector invids;
    Invid userInvid;
    String userName;

    StringBuffer result = new StringBuffer();

    /* -- */

    groupname = (String) object.getFieldValueLocal(groupSchema.GROUPNAME);

    // currently in the Ganymede schema, group passwords aren't in passfields.

    pass = (String) object.getFieldValueLocal(groupSchema.PASSWORD);

    Integer gidInt = (Integer) object.getFieldValueLocal(groupSchema.GID);

    if (gidInt == null)
      {
	Ganymede.debug("Error, couldn't find gid for group " + groupname);
	return;
      }

    gid = gidInt.intValue();
    
    // we currently don't explicitly record the home group.. just take the first group
    // that the user is in.

    invids = object.getFieldValuesLocal(groupSchema.USERS);

    if (invids == null)
      {
	// Ganymede.debug("BSDBuilder.writeUserLine(): null user list for group " + groupname);
      }
    else
      {
	for (int i = 0; i < invids.size(); i++)
	  {
	    userInvid = (Invid) invids.elementAt(i);
	    
	    userName = getLabel(userInvid);

	    if (userName != null)
	      {
		users.addElement(userName);
	      }
	  }
      }

    // now build our output line

    result.append(groupname);
    result.append(":");
    result.append(pass);
    result.append(":");
    result.append(Integer.toString(gid));
    result.append(":");

    for (int i = 0; i < users.size(); i++)
      {
	if (i > 0)
	  {
	    result.append(",");
	  }

	result.append((String) users.elementAt(i));
      }

    if (result.length() > 1024)
      {
	Ganymede.debug("BSDBuilderTask.writeGroupLine(): Warning!  group " + 
		       groupname + " overflows the BSD line length!");
      }

    writer.println(result.toString());
  }
}
