/*

   NTSambaBuilderTask.java

   This class is intended to dump the Ganymede datastore to a file
   that can be parsed by scripts that will update an NT domain
   controller and/or a Samba server.
   
   Created: 1 December 2000

   Last Mod Date: $Date$
   Last Revision Changed: $Rev$
   Last Changed By: $Author$
   SVN URL: $HeadURL$

   Module By: Jonathan Abbey, jonabbey@arlut.utexas.edu

   -----------------------------------------------------------------------
	    
   Ganymede Directory Management System
 
   Copyright (C) 1996-2005
   The University of Texas at Austin.

   Contact information

   Author Email: ganymede_author@arlut.utexas.edu
   Email mailing list: ganymede@arlut.utexas.edu

   US Mail:

   Computer Science Division
   Applied Research Laboratories
   The University of Texas at Austin
   PO Box 8029, Austin TX 78713-8029

   Telephone: (512) 835-3200

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA

*/

package arlut.csd.ganymede.userKit;

import arlut.csd.ganymede.server.*;
import arlut.csd.ganymede.common.*;
import arlut.csd.Util.PathComplete;
import arlut.csd.Util.FileOps;

import java.util.*;
import java.io.*;

/*------------------------------------------------------------------------------
                                                                           class
                                                              NTSambaBuilderTask

------------------------------------------------------------------------------*/

/**
 * This class is intended to dump the Ganymede datastore to a file
 * that can be parsed by scripts that will update an NT domain
 * controller and/or a Samba server.
 *
 * @author Jonathan Abbey jonabbey@arlut.utexas.edu
 *
 */

public class NTSambaBuilderTask extends GanymedeBuilderTask {

  private String buildScript = null;
  private String ntPasswdFile = null;
  private String ntGroupFile = null;
  private String sambaFileVersion1 = null;
  private String sambaFileVersion2 = null;

  // ---

  private Date now = null;
  private boolean backedup = false;

  /* -- */

  /**
   * <p>The constructor for GanymedeBuilderTask subclasses takes an
   * Invid for the DBObject in the Ganymede data store that defines
   * the task.  The taskDefObjInvid member variable is comes from the
   * GanymedeBuilderTask that we are inheriting from.  Setting it
   * in this constructor allows us to use the getOptionValue() method to read
   * our configuration strings for this task.</p>
   */

  public NTSambaBuilderTask(Invid _taskObjInvid)
  {
    // set the taskDefObjInvid in GanymedeBuilderTask so
    // we can look up option strings

    taskDefObjInvid = _taskObjInvid;
  }

  /**
   *
   * This method runs with a dumpLock obtained for the builder task.
   *
   * Code run in builderPhase1() can call enumerateObjects() and
   * baseChanged().
   *
   * @return true if builderPhase1 made changes necessitating the
   * execution of builderPhase2.
   *
   */

  public boolean builderPhase1()
  {
    PrintWriter out;
    boolean result = false;

    /* -- */

    backedup = false;
    now = null;

    // now see what our current option values are

    ntPasswdFile = getOptionValue("NTPasswdFile");
    ntGroupFile = getOptionValue("NTGroupFile");
    
    sambaFileVersion1 = getOptionValue("SambaVersion1");
    sambaFileVersion2 = getOptionValue("SambaVersion2");

    if (baseChanged(SchemaConstants.UserBase) || baseChanged((short) 257))
      {
	if (baseChanged(SchemaConstants.UserBase))
	  {
	    Ganymede.debug("NTSambaBuilderTask: Need to build user map");
	  }

	if (baseChanged((short) 257))
	  {
	    Ganymede.debug("NTSambaBuilderTask: Need to build group map");
	  }
      }

    if ((ntPasswdFile != null && !ntPasswdFile.equals("")) ||
	(ntGroupFile != null && !ntGroupFile.equals("")))
      {
	if (writeNTFiles())
	  {
	    result = true;
	  }
      }

    if (sambaFileVersion1 != null && !sambaFileVersion1.equals(""))
      {
	if (writeSambafileVersion1(sambaFileVersion1))
	  {
	    result = true;
	  }
      }

    if (sambaFileVersion2 != null && !sambaFileVersion2.equals(""))
      {
	if (writeSambafileVersion2(sambaFileVersion2))
	  {
	    result = true;
	  }
      }

    return result;
  }

  /**
   *
   * This method runs after this task's dumpLock has been
   * relinquished.  This method is intended to be used to finish off a
   * build process by running (probably external) code that does not
   * require direct access to the database.
   *
   * builderPhase2 is only run if builderPhase1 returns true.
   *
   */

  public boolean builderPhase2()
  {
    File
      file;

    /* -- */

    buildScript = getOptionValue("buildScript");

    if (buildScript == null || buildScript.equals(""))
      {
	Ganymede.debug("NTSambaBuilderTask: no buildScript defined in task object's option string vector");
	Ganymede.debug("                    not running any external update script.");
	return false;
      }

    file = new File(buildScript);

    if (file.exists())
      {
	Ganymede.debug("NTSambaBuilderTask builderPhase2 running");

	try
	  {
	    FileOps.runProcess(buildScript);
	  }
	catch (IOException ex)
	  {
	    Ganymede.debug("Couldn't exec buildScript (" + buildScript + ") due to IOException: " + ex);
	  }
	catch (InterruptedException ex)
	  {
	    Ganymede.debug("Failure during exec of buildScript (" + buildScript + "): " + ex);
	  }
      }
    else
      {
	Ganymede.debug(buildScript + " doesn't exist, not running external NT/Samba build script");
      }

    Ganymede.debug("NTSambaBuilderTask builderPhase2 completed");

    return true;
  }

  // ***
  //
  // The following private methods are used to support the builder logic.
  //
  // ***


  /**
   * <P>This method generates files that can be used to synchronize
   * passwords and accounts to an NT PDC and to Samba, and the like.</P>
   *
   * <P>This method writes out two files, 'ntPasswd', and 'ntGroup', which
   * contain all information necessary to synchronize user and group accounts
   * into NT and Samba.</P>
   *
   * <P>These files are basically equivalent to the standard UNIX versions
   * of these files, except that passwords are saved in plaintext form, and
   * that information not essential to NT and Samba (like the home directory
   * and shell) are excluded.</P>
   *
   * <P>It is expected that the external ntsambaBuild script will do differential
   * comparisons between the new files we are writing out and the old copies
   * of those files, in order to handle user creation/deletion/and rename.</P>
   *
   * <P>At the present time, the external ntsambaBuild script will translate
   * the files we write out, and the old copies of those files, to a single
   * file, structured along the lines of a traditional Windows .INI file for
   * further processing.</P>
   *
   * <P>The ntPasswd file will contain lines that look like this:</P>
   * <PRE>
   * username:invid:Status:plaintextpassword:Full Name:Location:Gecos Info
   * </PRE>
   *
   * <P>The plaintextpassword field will be empty (no spaces between the
   * leading and trailing colons) if the plain text password could not
   * be discovered for that user.  The external build code will not
   * try to do anything with such users until we can get ahold of their
   * plaintext data.</P>
   *
   * <P>And the ntGroup file will look like:</P>
   * <PRE>
   * groupname:invid:Status:username,username,username,username:Comment
   * </PRE>
   *
   * <P>The group Comment field provides a string that identifies this
   * group as having originated in Ganymede, and it identifies the owner
   * group (if any) that owns it.  The NT update code compares an existing
   * group's comment field and will refuse to touch it if the group wasn't
   * created by Ganymede.  This keeps things safe if some joker tries
   * to create an 'Admin' group in Ganymede.</P>
   *
   * <P>In both cases, the invid is the Ganymede object identifier from the
   * source number in Ganymede, and will look like '3,101', with the comma
   * used to separate the object type from the object identifier number.</P>
   *
   * <P>The Invid is included so that the external build process can reliably
   * track user and group renaming.</P>
   *
   * <P>The 'Status' field will be either 'active' or 'inactive', and is there
   * to allow us to temporarily turn accounts off in NT without removing them.</P>
   */

  private boolean writeNTFiles()
  {
    PrintWriter fileOut = null;
    DBObject user;
    DBObject group;
    Enumeration users;
    Enumeration groups;
    String status;

    /* -- */

    if (ntPasswdFile == null || ntPasswdFile.equals(""))
      {
	Ganymede.debug("NTSambaBuilderTask.writeNTFiles(): no ntPasswdFile option string set");
	return false;
      }

    if (ntGroupFile == null || ntGroupFile.equals(""))
      {
	Ganymede.debug("NTSambaBuilderTask.writeNTFiles(): no ntGroupFile option string set");
	return false;
      }

    try
      {
	fileOut = openOutFile(ntPasswdFile, "nt");
      }
    catch (IOException ex)
      {
	ex.printStackTrace();

	Ganymede.debug("NTSambaBuilderTask.writeNTFiles(): couldn't open " + ntPasswdFile + ": " + ex);
	return false;
      }

    try
      {
	users = enumerateObjects(SchemaConstants.UserBase);

	while (users.hasMoreElements())
	  {
	    String password;

	    /* -- */

	    user = (DBObject) users.nextElement();

	    if (user.isInactivated())
	      {
		status = "inactive";
	      }
	    else
	      {
		status = "active";
	      }

	    PasswordDBField passField = (PasswordDBField) user.getField(SchemaConstants.UserPassword);

	    if (passField == null)
	      {
		password = "";
	      }
	    else
	      {
		password = passField.getPlainText();

		if (password == null)
		  {
		    password = "";
		  }
	      }

	    // ok, write it.

	    fileOut.print(escapeString(user.getLabel()));
	    fileOut.print(":");
	    fileOut.print(user.getInvid().getType());
	    fileOut.print(",");
	    fileOut.print(user.getInvid().getNum());
	    fileOut.print(":");
	    fileOut.print(status);
	    fileOut.print(":");
	    fileOut.print(escapeString(password));
	    fileOut.print(":");

	    // make up a description for NT to show in its user info field

	    String fullname = (String) user.getFieldValueLocal(userSchema.FULLNAME);

	    StringBuffer compositeGecos = new StringBuffer();

	    String room = (String) user.getFieldValueLocal(userSchema.ROOM);

	    if (room != null)
	      {
		compositeGecos.append(room);
	      }

	    String workphone = (String) user.getFieldValueLocal(userSchema.OFFICEPHONE);

	    if (workphone != null)
	      {
		compositeGecos.append(",");

		compositeGecos.append(workphone);
	      }

	    String homephone = (String) user.getFieldValueLocal(userSchema.HOMEPHONE);

	    if (homephone != null)
	      {
		compositeGecos.append(",");

		compositeGecos.append(homephone);
	      }

	    fileOut.print(escapeString(fullname));
	    fileOut.print(":");
	    fileOut.println(escapeString(compositeGecos.toString()));
	  }
      }
    finally
      {
	fileOut.close();
      }

    try
      {
	fileOut = openOutFile(ntGroupFile, "nt");
      }
    catch (IOException ex)
      {
	ex.printStackTrace();

	Ganymede.debug("NTSambaBuilderTask.writeNTFiles(): couldn't open " + ntGroupFile + ": " + ex);
	return false;
      }

    try
      {
	groups = enumerateObjects((short) 257);

	while (groups.hasMoreElements())
	  {
	    group = (DBObject) groups.nextElement();

	    if (group.isInactivated())
	      {
		status = "inactive";
	      }
	    else
	      {
		status = "active";
	      }

	    fileOut.print(escapeString(group.getLabel()));
	    fileOut.print(":");
	    fileOut.print(group.getInvid().getType());
	    fileOut.print(",");
	    fileOut.print(group.getInvid().getNum());
	    fileOut.print(":");
	    fileOut.print(status);
	    fileOut.print(":");
	    
	    InvidDBField usersField = (InvidDBField) group.getField(groupSchema.USERS);

	    if (usersField != null)
	      {
		fileOut.print(escapeString(usersField.getValueString()));
	      }

	    fileOut.print(":Ganymede");

	    InvidDBField ownerField = (InvidDBField) group.getField(SchemaConstants.OwnerListField);

	    if (ownerField != null)
	      {
		fileOut.print(" [");
		fileOut.print(escapeString(ownerField.getValueString()));
		fileOut.print("]");
	      }

	    fileOut.println();
	  }
      }
    finally
      {
	fileOut.close();
      }

    return true;
  }

  /** 
   * We can't have any : characters in passwords in the rshNT.txt
   * file we generate, since we use : chars as field separators in
   * this file.  Make sure that we backslash any such chars.
   */
  
  private String escapeString(String in)
  {
    if (in == null || in.equals(""))
      {
	return "";
      }
    
    StringBuffer buffer = new StringBuffer();
    char[] ary = in.toCharArray();

    /* -- */

    // do it

    for (int i = 0; i < ary.length; i++)
      {
	if (ary[i] == ':')
	  {
	    buffer.append("\\:");
	  }
	else if (ary[i] == '\\')
	  {
	    buffer.append("\\\\");
	  }
	else
	  {
	    buffer.append(ary[i]);
	  }
      }

    return buffer.toString();
  }

  /**
   * <p>Samba version 1:</p>
   *
   * <p>broccol:12003:612EE67D1EFC2FB60B42BCD4578197DF:27A4F1E1E377CAD237C95B6146457F86:Jonathan Abbey,S321 CSD,3199,6803522:/home/broccol:/bin/tcsh</p>
   *
   */

  private boolean writeSambafileVersion1(String sambaFilename)
  {
    PrintWriter sambaFile = null;

    try
      {
	sambaFile = openOutFile(sambaFilename, "NTSambaBuilder");
      }
    catch (IOException ex)
      {
	System.err.println("GASHBuilderTask.writeSambaFileVersion1(): couldn't open " + 
			   sambaFilename + ": " + ex);
	return false;
      }

    try
      {
	Enumeration users = enumerateObjects(SchemaConstants.UserBase);

	while (users.hasMoreElements())
	  {
	    DBObject user = (DBObject) users.nextElement();

	    if (user.isInactivated())
	      {
		// we just leave inactivated users out of a Version 1
		// Samba password file

		continue;
	      }

	    String username = (String) user.getFieldValueLocal(userSchema.USERNAME);

	    PasswordDBField passField = (PasswordDBField) user.getField(userSchema.PASSWORD);

	    if (passField == null)
	      {
		continue;
	      }
	    
	    String hash1 = passField.getLANMANCryptText();

	    if (hash1 == null || hash1.equals(""))
	      {
		continue;
	      }

	    String hash2 = passField.getNTUNICODECryptText();

	    if (hash2 == null || hash2.equals(""))
	      {
		continue;
	      }

	    Integer uid = (Integer) user.getFieldValueLocal(userSchema.UID);

	    if (uid == null)
	      {
		continue;
	      }

	    String fullname = cleanString((String) user.getFieldValueLocal(userSchema.FULLNAME));
	    String room = cleanString((String) user.getFieldValueLocal(userSchema.ROOM));
	    String workphone = cleanString((String) user.getFieldValueLocal(userSchema.OFFICEPHONE));
	    String homephone = cleanString((String) user.getFieldValueLocal(userSchema.HOMEPHONE));
	    String homedir = cleanString((String) user.getFieldValueLocal(userSchema.HOMEDIR));
	    String shell = cleanString((String) user.getFieldValueLocal(userSchema.LOGINSHELL));
	    String composite = cleanString(fullname + "," + 
					   room + "," + 
					   workphone + "," + homephone);

	    sambaFile.println(username + ":" + uid.intValue() + ":" + 
			      hash1 + ":" + hash2 + ":" +
			      composite + ":" + homedir + ":" + shell);
	  }
      }
    finally
      {
	sambaFile.close();
      }

    return true;
  }

  /**
   * <p>Samba version 2:</p>
   *
   * <p>broccol:12003:612EE67D1EFC2FB60B42BCD4578197DF:27A4F1E1E377CAD237C95B6146457F86:[U          ]:LCT-375412BE:</p>
   */

  private boolean writeSambafileVersion2(String sambaFilename)
  {
    String hash1 = null;
    String hash2 = null;

    PrintWriter sambaFile = null;

    try
      {
	sambaFile = openOutFile(sambaFilename, "NTSambaBuilder");
      }
    catch (IOException ex)
      {
	System.err.println("GASHBuilderTask.writeSambaFileVersion2(): couldn't open " + 
			   sambaFilename + ": " + ex);
	return false;
      }

    try
      {
	Enumeration users = enumerateObjects(SchemaConstants.UserBase);

	while (users.hasMoreElements())
	  {
	    DBObject user = (DBObject) users.nextElement();

	    boolean inactivated = user.isInactivated();
	    
	    String username = (String) user.getFieldValueLocal(userSchema.USERNAME);

	    if (username == null || username.equals(""))
	      {
		continue;
	      }

	    PasswordDBField passField = (PasswordDBField) user.getField(userSchema.PASSWORD);

	    if (passField == null)
	      {
		inactivated = true;
	      }
	    
	    if (!inactivated)
	      {
		hash1 = passField.getLANMANCryptText();

		if (hash1 == null || hash1.equals(""))
		  {
		    inactivated = true;
		  }
		else
		  {
		    hash2 = passField.getNTUNICODECryptText();
		    
		    if (hash2 == null || hash2.equals(""))
		      {
			inactivated = true;
		      }
		  }
	      }

	    if (inactivated)
	      {
		hash1 = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
		hash2 = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
	      }

	    Integer uid = (Integer) user.getFieldValueLocal(userSchema.UID);

	    if (uid == null)
	      {
		continue;
	      }

	    // Samba 2.0 uses a flag string with 11 spaces and/or flag chars
	    // between a pair of brackets.

	    String flagString;

	    if (inactivated)
	      {
		flagString = "[UD         ]";
	      }
	    else
	      {
		flagString = "[U          ]";
	      }

	    // sanity checking

	    if (hash1 == null || hash1.length() != 32)
	      {
		throw new RuntimeException("bad LANMAN hash string: " + hash1);
	      }

	    if (hash2 == null || hash2.length() != 32)
	      {
		throw new RuntimeException("bad LANMAN hash string: " + hash1);
	      }

	    if (flagString.length() != 13)
	      {
		throw new RuntimeException("bad flag string");
	      }

	    // we are not currently keeping a separate timestamp for
	    // the password field, so tell Samba the password was
	    // changed at the last modification time for the user
	    // object as a whole.

	    Date modDate = (Date) user.getFieldValueLocal(SchemaConstants.ModificationDateField);

	    // modDate shouldn't ever be null, since objects should
	    // not be able to be checked into the database without
	    // having the ModificationDateField set.  If I'm ever
	    // wrong about this, the NullPointerException will have to
	    // set me straight.

	    String dateString = "LCT-" + dateToSMBHex(modDate.getTime());

	    sambaFile.println(username + ":" + uid.intValue() + ":" + 
			      hash1 + ":" + hash2 + ":" +
			      flagString + ":" + dateString);
	  }
      }
    finally
      {
	sambaFile.close();
      }

    return true;
  }

  /**
   * <p>Samba knows how to handle an 8 byte hex encoded date from the
   * version 2 smb.passwd file.  This method takes a standard long
   * Java timecode and generates an 8 byte hex string which holds the
   * number of seconds since epoch.</p>
   *
   * <p><b><blink>Note that this will overflow in the year 2038.</blink></b></p>
   */

  private String dateToSMBHex(long timecode)
  {
    timecode = timecode / 1000;

    if (timecode < 0)
      {
	throw new IllegalArgumentException("Time code is out of range from before the epoch");
      }

    if (timecode > java.lang.Integer.MAX_VALUE)
      {
	throw new IllegalArgumentException("Time code has overflowed");
      }

    StringBuffer timeString = new StringBuffer();

    timeString.append(java.lang.Integer.toHexString((int) timecode));

    // make sure we pad it out to 8 characters if it is less

    if (timeString.length() < 8)
      {
	for (int i = timeString.length(); i < 8; i++)
	  {
	    timeString.insert(0, "0");
	  }
      }

    return timeString.toString().toUpperCase();
  }

  /** 
   * We can't have any : characters in the Samba password file other than
   * as field separators, so we strip any we find out.
   */

  private String cleanString(String in)
  {
    if (in == null)
      {
	return "";
      }

    StringBuffer buffer = new StringBuffer();
    char[] ary = in.toCharArray();

    /* -- */

    // do it

    for (int i = 0; i < ary.length; i++)
      {
	if (ary[i] == ':')
	  {
	    continue;
	  }
	else
	  {
	    buffer.append(ary[i]);
	  }
      }

    return buffer.toString();
  }
}
