/*

   UNIXBuilderTask.java

   This class is intended to dump the Ganymede datastore to the
   UNIX passwd and group files.

   Created: 8 September 1998

   Module By: Jonathan Abbey, jonabbey@arlut.utexas.edu

   -----------------------------------------------------------------------

   Ganymede Directory Management System

   Copyright (C) 1996-2013
   The University of Texas at Austin.

   Ganymede is a registered trademark of The University of Texas at Austin

   Contact information

   Author Email: ganymede_author@arlut.utexas.edu
   Email mailing list: ganymede@arlut.utexas.edu

   US Mail:

   Computer Science Division
   Applied Research Laboratories
   The University of Texas at Austin
   PO Box 8029, Austin TX 78713-8029

   Telephone: (512) 835-3200

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA

*/

package arlut.csd.ganymede.userKit;

import arlut.csd.ganymede.server.*;
import arlut.csd.ganymede.common.*;
import arlut.csd.Util.PathComplete;
import arlut.csd.Util.FileOps;

import java.util.*;
import java.text.*;
import java.io.*;

/*------------------------------------------------------------------------------
                                                                           class
                                                                 UNIXBuilderTask

------------------------------------------------------------------------------*/

/**
 * <p>This class is intended to dump the Ganymede datastore to the
 * UNIX passwd and group files.</p>
 *
 * @author Jonathan Abbey jonabbey@arlut.utexas.edu
 */

public class UNIXBuilderTask extends GanymedeBuilderTask {

  private String buildScript = null;
  private String shacryptpasswdFile =  null;
  private String md5passwdFile =  null;
  private String passwdFile = null;
  private String groupFile = null;

  enum PasswdType
  {
    DESCRYPT,
    MD5CRYPT,
    SHACRYPT;
  }

  /* -- */

  /**
   * <p>The constructor for GanymedeBuilderTask subclasses takes an
   * Invid for the DBObject in the Ganymede data store that defines
   * the task.  The taskDefObjInvid member variable is comes from the
   * GanymedeBuilderTask that we are inheriting from.  Setting it
   * in this constructor allows us to use the getOptionValue() method to read
   * our configuration strings for this task.</p>
   */

  public UNIXBuilderTask(Invid _taskObjInvid)
  {
    // set the taskDefObjInvid in GanymedeBuilderTask so
    // we can look up option strings

    taskDefObjInvid = _taskObjInvid;
  }

  /**
   * <p>This method is intended to be overridden by subclasses of
   * GanymedeBuilderTask.</p>
   *
   * <p>This method runs with a dumpLock obtained for the builder
   * task.</p>
   *
   * <p>Code run in builderPhase1() can call enumerateObjects() and
   * baseChanged().  Note that the Enumeration of objects returned by
   * enumerateObjects() is only valid and should only be consulted
   * while builderPhase1 is running.. as soon as builderPhase1
   * returns, the dumpLock used to make the enumerateObjects() call
   * safe to use is relinquished, and any Enumerations obtained will
   * then be unsafe to depend on.</p>
   *
   * @return true if builderPhase1 made changes necessitating the
   * execution of builderPhase2.
   */

  public boolean builderPhase1()
  {
    PrintWriter out;
    boolean result = false;

    /* -- */

    // now see what our current option values are

    passwdFile = getOptionValue("passwdFile");
    md5passwdFile = getOptionValue("md5passwdFile");
    shacryptpasswdFile = getOptionValue("shacryptpasswdFile");

    if (passwdFile == null && md5passwdFile == null && shacryptpasswdFile == null)
      {
        Ganymede.debug("UNIXBuilderTask: error, no passwdFile specified");
      }

    if (passwdFile != null && passwdFile.equals(md5passwdFile))
      {
        Ganymede.debug("UNIXBuilderTask: error, md5passwdFile and passwdFile are the same, skipping md5passwdFile");
        md5passwdFile = null;
      }

    if (passwdFile != null && passwdFile.equals(shacryptpasswdFile))
      {
        Ganymede.debug("UNIXBuilderTask: error, shacryptpasswdFile and passwdFile are the same, skipping shacryptpasswdFile");
        shacryptpasswdFile = null;
      }

    groupFile = getOptionValue("groupFile");

    if (groupFile == null)
      {
        Ganymede.debug("UNIXBuilderTask: error, no groupFile specified");
      }

    // passwd

    if (baseChanged(SchemaConstants.UserBase))
      {
        Ganymede.debug("UNIXBuilderTask: Need to build user map");

        if (writePasswdFile(passwdFile, PasswdType.DESCRYPT))
          {
            result = true;
          }

        if (writePasswdFile(md5passwdFile, PasswdType.MD5CRYPT))
          {
            result = true;
          }

        if (writePasswdFile(shacryptpasswdFile, PasswdType.SHACRYPT))
          {
            result = true;
          }
      }

    // group

    if (baseChanged((short) 257))
      {
        Ganymede.debug("UNIXBuilderTask: Need to build group map");

        if (groupFile != null)
          {
            out = null;

            try
              {
                out = openOutFile(groupFile, "unix");
              }
            catch (IOException ex)
              {
                ex.printStackTrace();
                Ganymede.debug("UNIXBuilderTask.builderPhase1(): couldn't open " + groupFile + ": " + ex);
              }

            if (out != null)
              {
                Enumeration groups = enumerateObjects((short) 257);

                while (groups.hasMoreElements())
                  {
                    DBObject group = (DBObject) groups.nextElement();

                    writeGroupLine(group, out);
                  }

                out.close();
              }

            result = true;
          }
      }

    return result;
  }

  /**
   * <p>This method is intended to be overridden by subclasses of
   * GanymedeBuilderTask.</p>
   *
   * <p>This method runs after this task's dumpLock has been
   * relinquished.  This method is intended to be used to finish off a
   * build process by running (probably external) code that does not
   * require direct access to the database.</p>
   *
   * <p>For instance, for an NIS builder task, builderPhase1() would scan
   * the Ganymede object store and write out NIS-compatible source
   * files.  builderPhase1() would return, the run() method drops the
   * dump lock so that other transactions can be committed, and then
   * builderPhase2() can be run to turn those on-disk files written by
   * builderPhase1() into NIS maps.  This generally involves executing
   * an external Makefile, which can take an indeterminate period of
   * time.</p>
   *
   * <p>It is important that any external process run by
   * builderPhase2() blocks until it is finished doing the build.  If
   * the external script tries to put itself into the background and
   * return early, the Ganymede server will conclude that the external
   * build has completely finished, and it will feel free to
   * immediately schedule this builder task again, which may mean that
   * the builderPhase1() method will overwrite files the backgrounded
   * external builder process is still using.</p>
   *
   * <p>Note as well that the Ganymede server makes no guarantee as to
   * what the environment variables or current working directory will
   * be set to when any external builder scripts are executed.  If
   * your external scripts depend on these things, you should make
   * sure that your external builder script sets them itself.</p>
   *
   * <p>By releasing the dumpLock before we get to that point, we
   * minimize contention for users of the system.</p>
   *
   * <p>As a result of having dropped the dumpLock, enumerateObjects()
   * cannot be called by this method.</p>
   *
   * <p>builderPhase2() can throw a ServiceNotFoundException or
   * ServiceFailedException to indicate to the admin console that the
   * build failed, or it can simply return a boolean result to
   * indicate the same in a less specific fashion.</p>
   *
   * <p>builderPhase2 is only run if builderPhase1 returns true.</p>
   */

  public boolean builderPhase2()
  {
    File
      file;

    /* -- */

    buildScript = getOptionValue("buildScript");

    if (buildScript == null || buildScript.equals(""))
      {
        Ganymede.debug("UNIXBuilderTask: no buildScript defined in task object's option string vector");
        Ganymede.debug("                 not running any external update script.");
        return false;
      }

    file = new File(buildScript);

    if (file.exists())
      {
        Ganymede.debug("UNIXBuilderTask builderPhase2 running");

        try
          {
            FileOps.runProcess(buildScript);
          }
        catch (IOException ex)
          {
            Ganymede.debug("Couldn't exec buildScript (" + buildScript + ") due to IOException: " + ex);
          }
        catch (InterruptedException ex)
          {
            Ganymede.debug("Failure during exec of buildScript (" + buildScript + "): " + ex);
          }
      }
    else
      {
        Ganymede.debug(buildScript + " doesn't exist, not running external UNIX build script");
      }

    Ganymede.debug("UNIXBuilderTask builderPhase2 completed");

    return true;
  }

  // ***
  //
  // The following private methods are used to support the UNIX builder logic.
  //
  // ***

  private boolean writePasswdFile(String filename, PasswdType passwdStyle)
  {
    if (filename == null || filename.equals(""))
      {
        return false;
      }

    PrintWriter out;

    try
      {
        out = openOutFile(filename, "unix");
      }
    catch (IOException ex)
      {
        return false;
      }

    try
      {
        Enumeration users = enumerateObjects(SchemaConstants.UserBase);

        while (users.hasMoreElements())
          {
            DBObject user = (DBObject) users.nextElement();

            writeUserLine(user, out, passwdStyle);
          }

        return true;
      }
    catch (Exception ex)
      {
        return false;
      }
    finally
      {
        try
          {
            out.close();
          }
        catch (Exception ex2)
          {
            return false;
          }
      }
  }

  /**
   * <p>This method writes out a line to the passwd UNIX source file.</p>
   *
   * <p>The lines in this file look like the following.</p>
   *
   * <p>broccol:393T6k3e/9/w2:12003:12010:Jonathan Abbey,S321 CSD,3199,8343915:/home/broccol:/bin/tcsh</p>
   *
   * @param object An object from the Ganymede user object base
   * @param writer The destination for this user line
   * @param passwdStyle Determines which hash variant is emitted into the file
   */

  private void writeUserLine(DBObject object, PrintWriter writer, PasswdType passwdStyle)
  {
    String username;
    String cryptedPass = null;
    int uid;
    int gid;
    String name;
    String room;
    String officePhone;
    String homePhone;
    String directory;
    String shell;

    PasswordDBField passField;
    Vector invids;
    Invid groupInvid;
    DBObject group;

    StringBuffer result = new StringBuffer();

    /* -- */

    username = (String) object.getFieldValueLocal(SchemaConstants.UserUserName);

    passField = (PasswordDBField) object.getField(SchemaConstants.UserPassword);

    if (passField != null)
      {
        switch (passwdStyle)
          {
          case DESCRYPT:
            cryptedPass = passField.getUNIXCryptText();
            break;

          case MD5CRYPT:
            cryptedPass = passField.getMD5CryptText();
            break;

          case SHACRYPT:
            cryptedPass = passField.getShaUnixCryptText();
            break;
          }

        // if we couldn't find what we wanted above, try the others

        if (cryptedPass == null)
          {
            cryptedPass = passField.getShaUnixCryptText();
          }

        if (cryptedPass == null)
          {
            cryptedPass = passField.getMD5CryptText();
          }

        if (cryptedPass == null)
          {
            cryptedPass = passField.getUNIXCryptText();
          }
      }
    else
      {
        Ganymede.debug("UNIXBuilderTask.writeUserLine(): null password for user " + username);
        cryptedPass = "**Nopass**";
      }

    uid = ((Integer) object.getFieldValueLocal(userSchema.UID)).intValue();

    // get the gid

    groupInvid = (Invid) object.getFieldValueLocal(userSchema.HOMEGROUP); // home group

    if (groupInvid == null)
      {
        Ganymede.debug("UNIXBuilderTask.writeUserLine(): null gid for user " + username);
        gid = -1;
      }
    else
      {
        group = getObject(groupInvid);

        if (group == null)
          {
            Ganymede.debug("UNIXBuilderTask.writeUserLine(): couldn't get reference to home group");

            gid = -1;
          }
        else
          {
            Integer gidInt = (Integer) group.getFieldValueLocal(groupSchema.GID);

            if (gidInt == null)
              {
                Ganymede.debug("UNIXBuilderTask.writeUserLine(): couldn't get gid value");

                gid = -1;
              }
            else
              {
                gid = gidInt.intValue();
              }
          }
      }

    name = (String) object.getFieldValueLocal(userSchema.FULLNAME);
    room = (String) object.getFieldValueLocal(userSchema.ROOM);
    officePhone = (String) object.getFieldValueLocal(userSchema.OFFICEPHONE);
    homePhone = (String) object.getFieldValueLocal(userSchema.HOMEPHONE);
    directory = (String) object.getFieldValueLocal(userSchema.HOMEDIR);
    shell = (String) object.getFieldValueLocal(userSchema.LOGINSHELL);

    // now build our output line

    result.append(username);
    result.append(":");
    result.append(cryptedPass);
    result.append(":");
    result.append(Integer.toString(uid));
    result.append(":");
    result.append(Integer.toString(gid));
    result.append(":");

    result.append(name);

    if (room != null && !room.equals(""))
      {
        result.append(",");
        result.append(room);
      }

    if (officePhone != null && !officePhone.equals(""))
      {
        result.append(",");
        result.append(officePhone);
      }

    if (homePhone != null && !homePhone.equals(""))
      {
        result.append(",");
        result.append(homePhone);
      }

    result.append(":");
    result.append(directory);
    result.append(":");
    result.append(shell);

    if (result.length() > 1024)
      {
        Ganymede.debug("UNIXBuilderTask.writeUserLine(): Warning!  user " +
                       username + " overflows the UNIX line length!");
      }

    writer.println(result.toString());
  }

  /**
   * <p>This method writes out a line to the group UNIX source file.</p>
   *
   * <p>The lines in this file look like the following.</p>
   *
   * <p>adgacc:ZzZz:4015:hammp,jgeorge,dd,doodle,dhoss,corbett,monk</p>
   *
   * @param object An object from the Ganymede user object base
   * @param writer The destination for this user line
   */

  private void writeGroupLine(DBObject object, PrintWriter writer)
  {
    String groupname;
    String pass;
    int gid;
    Vector users = new Vector();

    PasswordDBField passField;
    Vector invids;
    Invid userInvid;
    String userName;

    StringBuffer result = new StringBuffer();

    /* -- */

    groupname = (String) object.getFieldValueLocal(groupSchema.GROUPNAME);

    // currently in the Ganymede schema, group passwords aren't in passfields.

    pass = (String) object.getFieldValueLocal(groupSchema.PASSWORD);

    Integer gidInt = (Integer) object.getFieldValueLocal(groupSchema.GID);

    if (gidInt == null)
      {
        Ganymede.debug("Error, couldn't find gid for group " + groupname);
        return;
      }

    gid = gidInt.intValue();

    // we currently don't explicitly record the home group.. just take the first group
    // that the user is in.

    invids = object.getFieldValuesLocal(groupSchema.USERS);

    if (invids == null)
      {
        // Ganymede.debug("UNIXBuilder.writeUserLine(): null user list for group " + groupname);
      }
    else
      {
        for (int i = 0; i < invids.size(); i++)
          {
            userInvid = (Invid) invids.elementAt(i);

            userName = getLabel(userInvid);

            if (userName != null)
              {
                users.addElement(userName);
              }
          }
      }

    // now build our output line

    result.append(groupname);
    result.append(":");
    result.append(pass);
    result.append(":");
    result.append(Integer.toString(gid));
    result.append(":");

    for (int i = 0; i < users.size(); i++)
      {
        if (i > 0)
          {
            result.append(",");
          }

        result.append((String) users.elementAt(i));
      }

    if (result.length() > 1024)
      {
        Ganymede.debug("UNIXBuilderTask.writeGroupLine(): Warning!  group " +
                       groupname + " overflows the UNIX line length!");
      }

    writer.println(result.toString());
  }
}
