/*

   userCustom.java

   This file is a management class for user objects in Ganymede.

   Created: 30 July 1997

   Module By: Jonathan Abbey, jonabbey@arlut.utexas.edu

   -----------------------------------------------------------------------

   Ganymede Directory Management System

   Copyright (C) 1996-2013
   The University of Texas at Austin.

   Ganymede is a registered trademark of The University of Texas at Austin

   Contact information

   Author Email: ganymede_author@arlut.utexas.edu
   Email mailing list: ganymede@arlut.utexas.edu

   US Mail:

   Computer Science Division
   Applied Research Laboratories
   The University of Texas at Austin
   PO Box 8029, Austin TX 78713-8029

   Telephone: (512) 835-3200

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA

*/

package arlut.csd.ganymede.userKit;

import arlut.csd.ganymede.server.*;
import arlut.csd.ganymede.common.*;

import arlut.csd.JDialog.JDialogBuff;
import arlut.csd.Util.FileOps;
import arlut.csd.Util.StringUtils;

import java.io.*;
import java.util.*;
import java.rmi.*;

/*------------------------------------------------------------------------------
                                                                           class
                                                                      userCustom

------------------------------------------------------------------------------*/

/**
 * <p>This class is the custom plug-in to handle the user object type in the
 * Ganymede server.  It does special validations of operations on the user,
 * handles inactivation and reactivation logic, and generates Wizards as
 * needed.</p>
 *
 * <p>See the userSchema.java file for a list of field definitions that this
 * module expects to work with.</p>
 *
 * @see arlut.csd.ganymede.userKit.userSchema
 * @see arlut.csd.ganymede.server.DBEditObject
 */

public class userCustom extends DBEditObject implements SchemaConstants, userSchema {

  static final boolean debug = false;
  static QueryResult shellChoices = new QueryResult();
  static Date shellChoiceStamp = null;

  static String homedir = null;
  static boolean externalsEnabled = false;
  static String newUserScript = null;
  static String deleteUserScript = null;
  static String renameUserScript = null;

  static Date schemaConfStamp = null;

  // ---

  QueryResult groupChoices = null;

  // remember the new username if it is being changed so our field chaining
  // will ok the change

  String newUsername;

  /**
   *
   * Customization Constructor
   *
   */

  public userCustom(DBObjectBase objectBase)
  {
    super(objectBase);
  }

  /**
   *
   * Create new object constructor
   *
   */

  public userCustom(DBObjectBase objectBase, Invid invid, DBEditSet editset)
  {
    super(objectBase, invid, editset);
  }

  /**
   *
   * Check-out constructor, used by DBObject.createShadow()
   * to pull out an object for editing.
   *
   */

  public userCustom(DBObject original, DBEditSet editset)
  {
    super(original, editset);
  }

  /**
   * <p>Initializes a newly created DBEditObject.</p>
   *
   * <p>When this method is called, the DBEditObject has been created,
   * its ownership set, and all fields defined in the controlling
   * {@link arlut.csd.ganymede.server.DBObjectBase DBObjectBase} have
   * been instantiated without defined values.  If this DBEditObject
   * is an embedded type, it will have been linked into its parent
   * object before this method is called.</p>
   *
   * <p>This method is responsible for filling in any default values
   * that can be calculated from the {@link
   * arlut.csd.ganymede.server.DBSession DBSession} associated with
   * the editset defined in this DBEditObject.</p>
   *
   * <p>If initialization fails for some reason, initializeNewObject()
   * will return a ReturnVal with an error result..  If the owning
   * GanymedeSession is not in bulk-loading mode (i.e.,
   * GanymedeSession.enableOversight is true), {@link
   * arlut.csd.ganymede.server.DBSession#createDBObject(short,
   * arlut.csd.ganymede.common.Invid, java.util.Vector)
   * DBSession.createDBObject()} will checkpoint the transaction
   * before calling this method.  If this method returns a failure
   * code, the calling method will rollback the transaction.  This
   * method has no responsibility for undoing partial initialization,
   * the checkpoint/rollback logic will take care of that.</p>
   *
   * <p>If enableOversight is false, DBSession.createDBObject() will
   * not checkpoint the transaction status prior to calling
   * initializeNewObject(), so it is the responsibility of this method
   * to handle any checkpointing needed.</p>
   *
   * <p>This method should be overridden in subclasses.</p>
   *
   * @return A ReturnVal indicating success or failure.  May
   * be simply 'null' to indicate success if no feedback need
   * be provided.
   */

  public ReturnVal initializeNewObject()
  {
    ReturnVal retVal;

    /* -- */

    // we don't want to do initialization if we are bulk-loading.

    if (!getGSession().enableOversight)
      {
        return null;
      }

    // need to find a uid for this user

    NumericDBField numField = (NumericDBField) getField(UID);

    if (numField == null)
      {
        return Ganymede.createErrorDialog("User Initialization Failure",
                                          "Couldn't find the uid field.. schema problem?");
      }

    DBNameSpace namespace = numField.getNameSpace();

    if (namespace == null)
      {
        return Ganymede.createErrorDialog("User Initialization Failure",
                                          "Couldn't find the uid namespace.. schema problem?");
      }

    // now, find a uid.. unfortunately, we have to use immutable Integers here.. not
    // the most efficient at all.

    Integer uidVal = new Integer(1001);

    while (!namespace.reserve(editset, uidVal))
      {
        uidVal = new Integer(uidVal.intValue()+1);
      }

    // we use setValueLocal so we can set a value that the user can't edit.

    retVal = numField.setValueLocal(uidVal);

    return retVal;
  }

  /**
   * <p>Customization method to control whether a specified field
   * is required to be defined at commit time for a given object.</p>
   *
   * <p>To be overridden on necessity in DBEditObject subclasses.</p>
   *
   * <p>Note that this method will not be called if the controlling
   * GanymedeSession's enableOversight is turned off, as in
   * bulk loading.</p>
   *
   * <p>Note as well that the designated label field for objects are
   * always required, whatever this method returns, and that this
   * requirement holds without regard to the GanymedeSession's
   * enableOversight value.</p>
   *
   * <p><b>*PSEUDOSTATIC*</b></p>
   */

  public boolean fieldRequired(DBObject object, short fieldid)
  {
    switch (fieldid)
      {
      case userSchema.USERNAME:
      case userSchema.UID:
      case userSchema.LOGINSHELL:
      case userSchema.HOMEDIR:
      case userSchema.PASSWORD:
      case userSchema.HOMEGROUP:
        return true;
      }

    return false;
  }

  /**
   * <p>This method returns a key that can be used by the client
   * to cache the value returned by choices().  If the client
   * already has the key cached on the client side, it
   * can provide the choice list from its cache rather than
   * calling choices() on this object again.</p>
   *
   * <p>The default logic in this method is designed to cause the client
   * to cache choice lists for invid fields in the 'all objects of
   * invid target type' cache bucket.  If your InvidDBField needs to
   * provide a restricted subset of objects of the targeted type as
   * the choice list, you'll need to override this method to either
   * return null (to turn off choice list caching), or generate some
   * kind of unique key that won't collide with the Short objects used
   * to represent the default object list caches.</p>
   *
   * <p>See also the {@link
   * arlut.csd.ganymede.server.DBEditObject#choiceListHasExceptions(arlut.csd.ganymede.server.DBField)}
   * hook, which controls whether or not the default logic will
   * encourage the client to cache a given InvidDBField's choice list.</p>
   *
   * <p>If there is no caching key, this method will return null.</p>
   *
   * <p>--------------------</p>
   *
   * <p>We don't want the HOMEGROUP field's choice list to be cached on
   * the client because it is dynamically generated for this
   * context, and doesn't make sense in other contexts.</p>
   */

  public Object obtainChoicesKey(DBField field)
  {
    if (field.getID() == HOMEGROUP)
      {
        return null;
      }
    else
      {
        return super.obtainChoicesKey(field);
      }
  }

  /**
   * <p>This method provides a hook that can be used to generate
   * choice lists for invid and string fields that provide
   * such.  String and Invid DBFields will call their owner's
   * obtainChoiceList() method to get a list of valid choices.</p>
   *
   * <p>This method will provide a reasonable default for targetted
   * invid fields, filtered by the GanymedeSession's
   * visibilityFilterInvids list.</p>
   *
   * <p>NOTE: This method does not need to be synchronized.  Making this
   * synchronized can lead to DBEditObject/DBSession nested monitor
   * deadlocks.</p>
   */

  public synchronized QueryResult obtainChoiceList(DBField field)
  {
    switch (field.getID())
      {
      case LOGINSHELL:                  // login shell

        updateShellChoiceList();

        if (debug)
          {
            System.err.println("userCustom: obtainChoice returning " + shellChoices + " for shell field.");
          }

        return shellChoices;

      case HOMEGROUP:                   // home group

        updateGroupChoiceList();
        return groupChoices;

      default:
        try
          {
            return super.obtainChoiceList(field);
          }
        catch (NotLoggedInException ex)
          {
            System.err.println("userCustom: obtainChoiceList() received NotLoggedInException: " + ex.getMessage());
            return null;
          }
      }
  }

  void updateGroupChoiceList()
  {
    if (groupChoices == null)
      {
        groupChoices = new QueryResult();

        Vector<Invid> invids = (Vector<Invid>) getFieldValuesLocal(GROUPLIST); // groups list

        for (Invid invid: invids)
          {
            groupChoices.addRow(invid, getLabel(invid), true); // must be editable because the client cares
          }
      }
  }

  void updateShellChoiceList()
  {
    synchronized (shellChoices)
      {
        DBObjectBase base = Ganymede.db.getObjectBase("Shell Choice");

        // just go ahead and throw the null pointer if we didn't get our base.

        if (shellChoiceStamp == null || shellChoiceStamp.before(base.getTimeStamp()))
          {
            System.err.println("userCustom - updateShellChoiceList()");

            shellChoices = new QueryResult();

            Query query = new Query("Shell Choice", null, false);

            // internalQuery doesn't care if the query has its filtered bit set

            System.err.println("userCustom - issuing query");

            Vector<Result> results = internalSession().internalQuery(query);

            System.err.println("userCustom - processing query results");

            for (Result result: results)
              {
                shellChoices.addRow(null, result.toString(), false); // no invid
              }

            if (shellChoiceStamp == null)
              {
                shellChoiceStamp = new Date();
              }
            else
              {
                shellChoiceStamp.setTime(System.currentTimeMillis());
              }
          }
      }
  }

  /**
   * <p>Customization method to verify whether the user has permission
   * to remove a given object.  The client's DBSession object
   * will call this per-class method to do an object type-
   * sensitive check to see if this object feels like being
   * available for removal by the client.</p>
   *
   * <p>To be overridden on necessity in DBEditObject subclasses.</p>
   *
   * <p><b>*PSEUDOSTATIC*</b></p>
   *
   * @return A ReturnVal indicating success or failure.  May
   * be simply 'null' to indicate success if no feedback need
   * be provided.
   */

  public ReturnVal canRemove(DBSession session, DBObject object)
  {
    String name = (String) object.getFieldValueLocal(userSchema.USERNAME);

    if (name != null && name.equals("root"))
      {
        return Ganymede.createErrorDialog("User Removal Error",
                                          "Deleting the root account may cause problems on your network." +
                                          " If you are sure you don't want this account in Ganymede, " +
                                          "rename it and then delete it.");
      }

    return super.canRemove(session,object);
  }

  /**
   * <p>This method allows the DBEditObject to have executive approval
   * of any scalar set operation, and to take any special actions in
   * reaction to the set.  When a scalar field has its value set, it
   * will call its owners finalizeSetValue() method, passing itself as
   * the &lt;field&gt; parameter, and passing the new value to be
   * approved as the &lt;value&gt; parameter.  A Ganymede customizer
   * who creates custom subclasses of the DBEditObject class can
   * override the finalizeSetValue() method and write his own logic
   * to examine any change and either approve or reject the change.</p>
   *
   * <p>A custom finalizeSetValue() method will typically need to
   * examine the field parameter to see which field is being changed,
   * and then do the appropriate checking based on the value
   * parameter.  The finalizeSetValue() method can call the normal
   * this.getFieldValueLocal() type calls to examine the current state
   * of the object, if such information is necessary to make
   * appropriate decisions.</p>
   *
   * <p>If finalizeSetValue() returns null or a ReturnVal object with
   * a positive success value, the DBField that called us is
   * guaranteed to proceed to make the change to its value.  If this
   * method returns a non-success code in its ReturnVal, as with the
   * result of a call to Ganymede.createErrorDialog(), the DBField
   * that called us will not make the change, and the field will be
   * left unchanged.  Any error dialog returned from finalizeSetValue()
   * will be passed to the user.</p>
   *
   * <p>The DBField that called us will take care of all standard
   * checks on the operation (including a call to our own
   * verifyNewValue() method before calling this method.  Under normal
   * circumstances, we won't need to do anything here.
   * finalizeSetValue() is useful when you need to do unusually
   * involved checks, and for when you want a chance to trigger other
   * changes in response to a particular field's value being
   * changed.</p>
   *
   * @return A ReturnVal indicating success or failure.  May
   * be simply 'null' to indicate success if no feedback need
   * be provided.
   */

  public synchronized ReturnVal finalizeSetValue(DBField field, Object value)
  {
    DBSession session = editset.getDBSession();
    StringDBField sf;

    /* -- */

    if (field.getID() == HOMEDIR)
      {
        String dir = (String) value;

        /* -- */

        sf = (StringDBField) getField(USERNAME);

        return null;
      }

    if (field.getID() == USERNAME)
      {
        try
          {
            newUsername = (String) value;

            // if we are being told to clear the user name field, go ahead and
            // do it.. we assume this is being done by user removal logic,
            // so we won't force everything to go through a wizard.

            if (isDeleting() && (value == null))
              {
                return null;
              }

            // update the home directory location.. we assume that if
            // the user has permission to rename the user, they can
            // automatically execute this change to the home directory.

            readPrefs();

            if (homedir != null)
              {
                sf = (StringDBField) getField(HOMEDIR);
                sf.setValueLocal(homedir + (String) value);
              }

            // rename all the associated persona with the new user name

            InvidDBField inv = (InvidDBField) getField(PERSONAE);

            if (inv == null)
              {
                return null;
              }

            sf = (StringDBField) getField(USERNAME); // old user name

            String oldName = (String) sf.getValueLocal();

            Vector<Invid> personaeInvids = (Vector<Invid>) inv.getValuesLocal();

            for (Invid invid: personaeInvids)
              {
                adminPersonaCustom adminObj = (adminPersonaCustom) session.editDBObject(invid);

                adminObj.refreshLabelField(null, null, (String) value);
              }
          }
        finally
          {
            newUsername = null;
          }
      }

    return null;
  }

  /**
   * <p>This method is the hook that DBEditObject subclasses use to interpose
   * {@link arlut.csd.ganymede.server.GanymediatorWizard wizards} when a field's
   * value is being changed.</p>
   *
   * <p>Whenever a field is changed in this object, this method will be
   * called with details about the change. This method can refuse to
   * perform the operation, it can make changes to other objects in
   * the database in response to the requested operation, or it can
   * choose to allow the operation to continue as requested.</p>
   *
   * <p>In the latter two cases, the wizardHook code may specify a list
   * of fields and/or objects that the client may need to update in
   * order to maintain a consistent view of the database.</p>
   *
   * <p>If server-local code has called
   * {@link arlut.csd.ganymede.server.GanymedeSession#enableOversight(boolean)
   * enableOversight(false)},
   * this method will never be
   * called.  This mode of operation is intended only for initial
   * bulk-loading of the database.</p>
   *
   * <p>This method may also be bypassed when server-side code uses
   * setValueLocal() and the like to make changes in the database.</p>
   *
   * <p>This method is called before the finalize*() methods.. the finalize*()
   * methods is where last minute cascading changes should be performed..
   * Note as well that wizardHook() is called before the namespace checking
   * for the proposed value is performed, while the finalize*() methods are
   * called after the namespace checking.</p>
   *
   * <p>The operation parameter will be a small integer, and should hold one of the
   * following values:</p>
   *
   * <dl>
   * <dt>1 - SETVAL</dt>
   * <dd>This operation is used whenever a simple scalar field is having
   * it's value set.  param1 will be the value being placed into the field.</dd>
   * <dt>2 - SETELEMENT</dt>
   * <dd>This operation is used whenever a value in a vector field is being
   * set.  param1 will be an Integer holding the element index, and
   * param2 will be the value being set.</dd>
   * <dt>3 - ADDELEMENT</dt>
   * <dd>This operation is used whenever a value is being added to the
   * end of a vector field.  param1 will be the value being added.</dd>
   * <dt>4 - DELELEMENT</dt>
   * <dd>This operation is used whenever a value in a vector field is being
   * deleted.  param1 will be an Integer holding the element index.</dd>
   * <dt>5 - ADDELEMENTS</dt>
   * <dd>This operation is used whenever a set of elements is being
   * added to a vector field en masse.  param1 will be a Vector containing
   * the values that are being added.</dd>
   * <dt>6 - DELELEMENTS</dt>
   * <dd>This operation is used whenever a set of elements is being
   * deleted from a vector field en masse.  param1 will be a Vector containing
   * the values that are being deleted.</dd>
   * <dt>7 - SETPASSPLAIN</dt>
   * <dd>This operation is used when a password field is having its password
   * set using a plaintext source.  param1 will be a String containing the
   * submitted password, or null if the password is being cleared.</dd>
   * <dt>8 - SETPASSCRYPT</dt>
   * <dd>This operation is used when a password field is having its password
   * set using a UNIX crypt() hashed source.  param1 will be a String containing the
   * submitted hashed password, or null if the password is being cleared.</dd>
   * <dt>9 - SETPASSMD5</dt>
   * <dd>This operation is used when a password field is having its password
   * set using an md5Ccrypt() hashed source.  param1 will be a String containing the
   * submitted hashed password, or null if the password is being cleared.</dd>
   * <dt>10 - SETPASSWINHASHES</dt>
   * <dd>This operation is used when a password field is having its password
   * set using Windows style password hashes.  param1 will be the password in
   * LANMAN hash form, param2 will be the password in NT Unicode MD4 hash
   * form.  Either or both of param1 and param2 may be null.</dd>
   * <dt>11 - SETPASSAPACHEMD5</dt>
   * <dd>This operation is used when a password field is having its
   * password set using the Apache variant of the md5crypt algorithm.
   * param1 will be the password in Apache md5crypt hash form, or null
   * if the password hash is being cleared.  param2 will be null.</dd>
   * <dt>12 - SETPASSSSHA</dt>
   * <dd>This operation is used when a password field is having its
   * password set using the OpenLDAP-style SSHA password hash.  param1
   * will be the password in SSHA form, or null if the password is
   * being cleared.  param2 will be null.</dd>
   * <dt>13 - SETPASS_SHAUNIXCRYPT</dt>
   * <dd>This operation is used when a password field is having its
   * password set using Ulrich Drepper's SHA256 or SHA512 Unix Crypt
   * algorithms.  param1 will be the password in SHA Unix Crypt form,
   * or null if the password is being cleared.  param2 will be
   * null.</dd>
   * <dt>14 - SETPASS_BCRYPT</dt>
   * <dd>This operation is used when a password field is having its
   * password set using the OpenBSD-style BCrypt password hash.  param1
   * will be the password in BCrypt form, or null if the password is
   * being cleared.  param2 will be null.</dd>
   * </dl>
   *
   * <p>To be overridden on necessity in DBEditObject subclasses.</p>
   *
   * @return null if the operation is approved without comment, or a
   * ReturnVal object indicating success or failure, objects and
   * fields to be rescanned by the client, and a doNormalProcessing
   * flag that will indicate to the field code whether or not the
   * operation should continue to completion using the field's
   * standard logic.  <b>It is very important that wizardHook return a
   * new ReturnVal(true, true) if the wizardHook wishes to simply
   * specify rescan information while having the field perform its
   * standard operation.</b> wizardHook() may return new
   * ReturnVal(true, false) if the wizardHook performs the operation
   * (or a logically related operation) itself.  The same holds true
   * for the respond() method in GanymediatorWizard subclasses.
   */

  public ReturnVal wizardHook(DBField field, int operation, Object param1, Object param2)
  {
    userHomeGroupDelWizard groupWizard = null;
    userRenameWizard renameWizard = null;
    ReturnVal result;

    /* -- */

    // if the groups field is being changed, we may need to intervene

    if (debug)
      {
        System.err.println("userCustom ** entering wizardHook, field = " + field.getName() + ", op= " + operation);
      }

    if (field.getID() == GROUPLIST)
      {
        switch (operation)
          {
          case ADDELEMENT:

            // ok, no big deal, but we will need to have the client
            // rescan the choice list for the home group field

            result = new ReturnVal(true, true);
            result.addRescanField(this.getInvid(), HOMEGROUP);
            groupChoices = null;
            return result;

          case DELELEMENT:

            if (isDeleting())
              {
                return null;
              }

            // ok, this is more of a big deal.. first, see if the value
            // being deleted is the home group.  If not, still no big
            // deal.

            int index = ((Integer) param1).intValue();

            Vector valueAry = getFieldValuesLocal(GROUPLIST);
            Invid delVal = (Invid) valueAry.elementAt(index);

            if (debug)
              {
                Ganymede.debug("userCustom: deleting group element " + getLabel(delVal));
              }

            if (!delVal.equals(getFieldValueLocal(HOMEGROUP)))
              {
                // whew, no big deal.. they are not removing the
                // home group.  The client will need to rescan,
                // but no biggie.

                if (debug)
                  {
                    Ganymede.debug("userCustom: I don't think " + getLabel(delVal) +
                                   " is the home group");
                  }

                result = new ReturnVal(true, true);
                result.addRescanField(this.getInvid(), HOMEGROUP);
                groupChoices = null;
                return result;
              }

            if (gSession.isWizardActive() && gSession.getWizard() instanceof userHomeGroupDelWizard)
              {
                groupWizard = (userHomeGroupDelWizard) gSession.getWizard();

                if (groupWizard.getState() == groupWizard.DONE)
                  {
                    // ok, assume the wizard has taken care of getting everything prepped and
                    // approved for us.  An active wizard has approved the operation

                    groupWizard.unregister();

                    return null;
                  }
                else
                  {
                    if (groupWizard.userObject != this)
                      {
                        Ganymede.debug("userCustom.wizardHook(): bad object");
                      }

                    if (groupWizard.getState() != groupWizard.DONE)
                      {
                        Ganymede.debug("userCustom.wizardHook(): bad state: " + groupWizard.getState());
                      }

                    groupWizard.unregister();

                    return Ganymede.createErrorDialog("User Object Error",
                                                      "The client is attempting to do an operation on " +
                                                      "a user object with an active wizard.");
                  }
              }
            else if (gSession.isWizardActive() && !(gSession.getWizard() instanceof userHomeGroupDelWizard))
              {
                return Ganymede.createErrorDialog("User Object Error",
                                                  "The client is attempting to do an operation on " +
                                                  "a user object with mismatched active wizard.");
              }

            // eek.  they are deleting the home group.  Why Lord, why?!

            try
              {
                groupWizard = new userHomeGroupDelWizard(this.gSession,
                                                         this,
                                                         param1);
              }
            catch (RemoteException ex)
              {
                throw new RuntimeException("Couldn't create userWizard " + ex.getMessage());
              }

            // if we get here, the wizard was able to register
            // itself.. go ahead and return the initial dialog for the
            // wizard.  The ReturnVal code that wizard.respond()
            // returns will have the success code set to false, so
            // whatever triggered us will prematurely exit, returning
            // the wizard's dialog.

            return groupWizard.respond(null);
          }
      }

    if ((field.getID() != USERNAME) ||
        (operation != SETVAL))
      {
        return null;            // by default, we just ok whatever else
      }

    // if we have not previously committed this user, we don't care if the
    // user's name is set.

    if ((field.getValueLocal() == null) || (getStatus() == ObjectStatus.CREATING))
      {
        result = new ReturnVal(true, true); // have setValue() do the right thing

        result.addRescanField(this.getInvid(), userSchema.HOMEDIR);

        return result;
      }

    if (!gSession.enableWizards)
      {
        return null;            // no wizards if the user is non-interactive.
      }

    // looks like we're renaming this user

    if (gSession.isWizardActive() && gSession.getWizard() instanceof userRenameWizard)
      {
        renameWizard = (userRenameWizard) gSession.getWizard();

        if ((renameWizard.getState() == renameWizard.DONE) &&
            (renameWizard.field == field) &&
            (renameWizard.userObject == this) &&
            (renameWizard.newname == param1))
          {
            // ok, assume the wizard has taken care of getting everything prepped and
            // approved for us.  An active wizard has approved the operation

            renameWizard.unregister();

            return null;
          }
        else
          {
            if (renameWizard.field != field)
              {
                Ganymede.debug("userCustom.wizardHook(): bad field");
              }

            if (renameWizard.userObject != this)
              {
                Ganymede.debug("userCustom.wizardHook(): bad object");
              }

            if (renameWizard.newname != param1)
              {
                Ganymede.debug("userCustom.wizardHook(): bad param");
              }

            if (renameWizard.getState() != renameWizard.DONE)
              {
                Ganymede.debug("userCustom.wizardHook(): bad state: " + renameWizard.getState());
              }

            renameWizard.unregister();
            return Ganymede.createErrorDialog("User Object Error",
                                              "The client is attempting to do an operation on " +
                                              "a user object with an active wizard.");
          }
      }
    else if (gSession.isWizardActive() && !(gSession.getWizard() instanceof userRenameWizard))
      {
        return Ganymede.createErrorDialog("User Object Error",
                                          "The client is attempting to do an operation on " +
                                          "a user object with mismatched active wizard.");
      }
    else
      {
        // there's no wizard active, and this operation has to be approved by one.  Go ahead
        // and set up the wizard and let the client play with it.

        // if we're setting the field to null, don't need to pass it through
        // a wizard.. we're probably just deleting this user.

        if (isDeleting() && (param1 == null))
          {
            return null;
          }

        try
          {
            renameWizard = new userRenameWizard(this.gSession,
                                                this,
                                                field,
                                                (String) param1);
          }
        catch (RemoteException ex)
          {
            throw new RuntimeException("Couldn't create userWizard " + ex.getMessage());
          }

        // if we get here, the wizard was able to register itself.. go ahead
        // and return the initial dialog for the wizard.  The ReturnVal code
        // that wizard.respond() returns will have the success code
        // set to false, so whatever triggered us will prematurely exit,
        // returning the wizard's dialog.

        return renameWizard.respond(null);
      }
  }

  /**
   * <p>This method is a hook for subclasses to override to
   * pass the phase-two commit command to external processes.</p>
   *
   * <p>For normal usage this method would not be overridden.  For
   * cases in which change to an object would result in an external
   * process being initiated whose <b>success or failure would not
   * affect the successful commit of this DBEditObject in the
   * Ganymede server</b>, the process invocation should be placed here,
   * rather than in
   * {@link arlut.csd.ganymede.server.DBEditObject#commitPhase1() commitPhase1()}.</p>
   *
   * <p>commitPhase2() is generally the last method called on a
   * DBEditObject before it is discarded by the server in the
   * {@link arlut.csd.ganymede.server.DBEditSet DBEditSet}
   * {@link arlut.csd.ganymede.server.DBEditSet#commit(java.lang.String) commit()} method.</p>
   *
   * <p>Subclasses that override this method may wish to make this method
   * synchronized.</p>
   *
   * <p><b>WARNING!</b> this method is called at a time when portions
   * of the database are locked for the transaction's integration into
   * the database.  You must not call methods that seek to gain a lock
   * on the Ganymede database.  At this point, this means no composite
   * queries on embedded object types, where you seek an object based
   * on a field in an embedded object and in the object itself, using
   * the GanymedeSession query calls, or else you will lock the server.</p>
   *
   * <p>This method should NEVER try to edit or change any DBEditObject
   * in the server.. at this point in the game, the server has fixed the
   * transaction working set and is depending on commitPhase2() not trying
   * to make changes internal to the server.</p>
   *
   * <p>To be overridden on necessity in DBEditObject subclasses.</p>
   */

  public void commitPhase2()
  {
    switch (getStatus())
      {
      case DROPPING:
        // the user never really existed.. no external actions required.
        break;

      case CREATING:

        // handle creating the user.. creating their home directory, setting
        // up their mail spool, etc., etc.

        createUserExternals();

        break;

      case DELETING:
        handleUserDelete(original.getLabel());
        break;

      case EDITING:

        // did the user's name change?

        String name = getLabel();
        String oldname = original.getLabel();

        if (!name.equals(oldname))
          {
            handleUserRename(oldname, name);
          }
      }

    return;
  }

  /**
   * <p>This method runs from userCustom's commitPhase2() and runs an
   * external script that can create the user's home directory, and
   * anything else that might need doing.</p>
   */

  private void createUserExternals()
  {
    if (debug)
      {
        System.err.println("userCustom: " + getLabel() + ", in createUserExternals().");
      }

    Integer id = (Integer) getFieldValueLocal(userSchema.UID);
    String homeDir = (String) getFieldValueLocal(userSchema.HOMEDIR);

    Invid homegroupInvid = (Invid) getFieldValueLocal(userSchema.HOMEGROUP);
    DBObject homeGroup = getDBSession().viewDBObject(homegroupInvid);
    Integer gid = (Integer) homeGroup.getFieldValueLocal(groupSchema.GID);

    if (debug)
      {
        System.err.println("userCustom: createUserExternals: getting newUserScript");
      }

    readPrefs();

    if (newUserScript == null)
      {
        if (externalsEnabled)
          {
            Ganymede.debug("userCustom.createUserExternals(): No new user script defined in schema configuration object");
          }

        return;
      }

    File createHandler = new File(newUserScript);

    if (createHandler.exists())
      {
        try
          {
            // we'll call our external script with the following
            //
            // parameters: <directory> <username> <user id> <group id>

            String execLine = newUserScript + " " + homeDir + " " + getLabel() + " " + id + " " + gid;

            if (debug)
              {
                System.err.println("createUserExternals: running " + execLine);
              }

            int result = FileOps.runProcess(execLine);

            if (result != 0)
              {
                Ganymede.debug("Couldn't handle externals for creating user " + getLabel() +
                               "\n" + newUserScript + " returned a non-zero result: " + result);
              }
          }
        catch (IOException ex)
          {
            Ganymede.debug("Couldn't handle externals for creating user " + getLabel() + "\n" +
                           ex.getMessage());
          }
        catch (InterruptedException ex)
          {
            Ganymede.debug("Couldn't handle externals for creating user " + getLabel() + "\n" +
                           ex.getMessage());
          }
      }
  }

  /**
   * This method handles external actions for renaming a user.
   */

  private void handleUserRename(String orig, String newname)
  {
    boolean success = false;

    /* -- */

    if (debug)
      {
        System.err.println("userCustom.handleUserRename(): user " + orig +
                           "is being renamed to " + newname);
      }

    try
      {
        readPrefs();

        if (renameUserScript == null)
          {
            if (externalsEnabled)
              {
                Ganymede.debug("userCustom.handleUserRename(): Couldn't find " +
                               "rename user script in schema configuration object");
              }

            return;
          }

        File renameHandler = new File(renameUserScript);

        if (renameHandler.exists())
          {
            try
              {
                String execLine = renameUserScript + " " + orig + " " + newname;

                if (debug)
                  {
                    System.err.println("handleUserRename: running " + execLine);
                  }

                int result = FileOps.runProcess(execLine);

                if (result != 0)
                  {
                    Ganymede.debug("Couldn't handle externals for renaming user " + orig +
                                   " to " + newname + "\n" + renameUserScript +
                                   " returned a non-zero result: " + result);
                  }
                else
                  {
                    success = true;
                  }
              }
            catch (IOException ex)
              {
                Ganymede.debug("Couldn't handle externals for renaming user " + orig +
                               " to " +
                               newname + "\n" +
                               ex.getMessage());
              }
            catch (InterruptedException ex)
              {
                Ganymede.debug("Couldn't handle externals for renaming user " + orig +
                               " to " +
                               newname + "\n" +
                               ex.getMessage());
              }
          }
      }
    finally
      {
        Invid admin = getGSession().getIdentityInvid();
        String adminName = getGSession().getIdentity();
        Vector<Invid> objects = new Vector<Invid>();
        objects.add(getInvid());

        StringBuffer buffer = new StringBuffer();

        buffer.append("User ");
        buffer.append(orig);
        buffer.append(" has been renamed to ");
        buffer.append(newname);
        buffer.append(".\n\n");

        if (success)
          {
            buffer.append("The user's main home directory has been renamed.  You may need ");
            buffer.append("to take some action to make sure that the user's account name change ");
            buffer.append("doesn't cause problems in your local scripts, etc.\n\n");
          }
        else
          {
            buffer.append("The user's main home directory was not able to be properly renamed ");
            buffer.append("by Ganymede.  You should contact a systems administrator on the user's");
            buffer.append("main server to make sure his or her home directory is renamed properly.\n\n");
            buffer.append("In addition, you may need ");
            buffer.append("to take some action to make sure that the user's account name change ");
            buffer.append("doesn't cause problems in your local scripts, etc.\n\n");
          }

        editset.logEvent("userrenamed",
                         buffer.toString(),
                         admin, adminName, objects, null);
      }
  }

  /**
   * This method handles external actions for deleting a user.
   */

  private void handleUserDelete(String name)
  {
    if (debug)
      {
        System.err.println("userCustom.handleUserDelete(): user " + name +
                           "is being deleted");
      }

    readPrefs();

    if (deleteUserScript == null)
      {
        if (externalsEnabled)
          {
            Ganymede.debug("userCustom.handleUserDelete(): No delete user script configured in schema configuration object");
          }

        return;
      }

    File deleteHandler = new File(deleteUserScript);

    if (deleteHandler.exists())
      {
        try
          {
            String execLine = deleteUserScript + " " + name;

            if (debug)
              {
                System.err.println("handleUserDelete: running " + execLine);
              }

            int result = FileOps.runProcess(execLine);

            if (result != 0)
              {
                Ganymede.debug("Couldn't handle externals for deleting user " + name +
                               "\n" + deleteUserScript +
                               " returned a non-zero result: " + result);
              }
          }
        catch (IOException ex)
          {
            Ganymede.debug("Couldn't handle externals for deleting user " + name +
                           ex.getMessage());
          }
        catch (InterruptedException ex)
          {
            Ganymede.debug("Couldn't handle externals for deleting user " + name +
                           ex.getMessage());
          }

        Invid admin = getGSession().getIdentityInvid();
        String adminName = getGSession().getIdentity();
        Vector objects = new Vector();
        objects.addElement(getInvid());

        StringBuffer buffer = new StringBuffer();

        buffer.append("User ");
        buffer.append(name);
        buffer.append(" has been expunged from the Ganymede database.\n\n");

        editset.logEvent("userdeleted",
                         buffer.toString(),
                         admin, adminName, objects, null);
      }
  }

  /**
   * <P>This method reads configuration information needed by the userCustom class from
   * the schema configuration object defined in userKit.</P>
   */

  public synchronized void readPrefs()
  {
    GanymedeSession mySession;
    boolean newSession = false;
    DBObjectBase base = Ganymede.db.getObjectBase((short) 256);

    if (schemaConfStamp == null || schemaConfStamp.before(base.getTimeStamp()))
      {
        mySession = getGSession();

        if (mySession == null)
          {
            // open a quick root session

            try
              {
                mySession = new GanymedeSession();
                newSession = true;
              }
            catch (RemoteException ex)
              {
                ex.printStackTrace();
                return;
              }
          }

        try
          {
            List<DBObject> objects = getDBSession().getTransactionalObjects((short) 256);    // schema configuration object is type 256

            if (objects == null || objects.size() != 1)
              {
                return;
              }

            DBObject obj =  objects.get(0);

            // field 256 is the home directory prefix

            homedir = (String) obj.getFieldValueLocal((short) 256);

            homedir = arlut.csd.Util.PathComplete.completePath(homedir);

            if (debug)
              {
                System.err.println("userCustom.readPrefs(): homedir = " + homedir);
              }

            // field 261 is the enable external actions checkbox

            Boolean externalsOk = (Boolean) obj.getFieldValueLocal((short) 261);

            if (externalsOk == null || !externalsOk.booleanValue())
              {
                newUserScript = null;
                deleteUserScript = null;
                renameUserScript = null;
                externalsEnabled = false;

                if (debug)
                  {
                    System.err.println("userCustom.readPrefs(): external scripts disabled");
                  }
              }
            else
              {
                externalsEnabled = true;

                // field 258 is the new user script

                newUserScript = (String) obj.getFieldValueLocal((short) 258);

                if (debug)
                  {
                    System.err.println("userCustom.readPrefs(): newUserScript = " + newUserScript);
                  }

                // field 259 is the delete user script

                deleteUserScript = (String) obj.getFieldValueLocal((short) 259);

                if (debug)
                  {
                    System.err.println("userCustom.readPrefs(): deleteUserScript = " + deleteUserScript);
                  }

                // field 260 is the rename user script

                renameUserScript = (String) obj.getFieldValueLocal((short) 260);

                if (debug)
                  {
                    System.err.println("userCustom.readPrefs(): renameUserScript = " + renameUserScript);
                  }
              }
          }
        finally
          {
            schemaConfStamp = new Date();

            if (newSession)
              {
                mySession.logout();
              }
          }
      }
  }

  /**
   * <p>Customization method to verify whether a specific field
   * in object should be cloned using the basic field-clone
   * logic.</p>
   *
   * <p>To be overridden in DBEditObject subclasses.</p>
   *
   * <p><b>*PSEUDOSTATIC*</b></p>
   */

  public boolean canCloneField(DBSession session, DBObject object, DBField field)
  {
    short fieldid = field.getID();

    switch (fieldid)
      {
      case USERNAME:
      case UID:
      case userSchema.PASSWORD:
      case HOMEDIR:
      case PERSONAE:
        return false;
      }

    return super.canCloneField(session, object, field);
  }
}
